package com.sun.org.apache.xalan.internal.xsltc.compiler;

import com.sun.org.apache.bcel.internal.generic.TargetLostException;
import com.sun.org.apache.bcel.internal.generic.InstructionHandle;
import com.sun.org.apache.bcel.internal.util.InstructionFinder;
import com.sun.org.apache.bcel.internal.generic.GETFIELD;
import com.sun.org.apache.bcel.internal.generic.NEW;
import com.sun.org.apache.bcel.internal.generic.INVOKESTATIC;
import com.sun.org.apache.bcel.internal.generic.LocalVariableGen;
import com.sun.org.apache.bcel.internal.generic.ISTORE;
import com.sun.org.apache.bcel.internal.generic.INVOKEINTERFACE;
import com.sun.org.apache.bcel.internal.generic.PUTFIELD;
import com.sun.org.apache.bcel.internal.generic.INVOKESPECIAL;
import com.sun.org.apache.bcel.internal.generic.ConstantPoolGen;
import com.sun.org.apache.bcel.internal.generic.INVOKEVIRTUAL;
import com.sun.org.apache.bcel.internal.generic.NEWARRAY;
import com.sun.org.apache.bcel.internal.generic.BasicType;
import com.sun.org.apache.bcel.internal.generic.GETSTATIC;
import com.sun.org.apache.bcel.internal.generic.PUTSTATIC;
import com.sun.org.apache.bcel.internal.generic.ANEWARRAY;
import com.sun.org.apache.bcel.internal.generic.PUSH;
import com.sun.org.apache.bcel.internal.generic.InstructionList;
import com.sun.org.apache.bcel.internal.generic.FieldGen;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.Util;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.MethodGenerator;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.TypeCheckError;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.Type;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.ClassGenerator;
import java.util.List;
import com.sun.org.apache.xalan.internal.xsltc.compiler.util.ErrorMsg;
import java.util.StringTokenizer;
import com.sun.org.apache.xml.internal.utils.SystemIDResolver;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Properties;
import java.util.Map;
import java.util.Vector;

public final class Stylesheet extends SyntaxTreeNode
{
    private String _version;
    private QName _name;
    private String _systemId;
    private Stylesheet _parentStylesheet;
    private Vector _globals;
    private Boolean _hasLocalParams;
    private String _className;
    private final Vector _templates;
    private Vector _allValidTemplates;
    private int _nextModeSerial;
    private final Map<String, Mode> _modes;
    private Mode _defaultMode;
    private final Map<String, String> _extensions;
    public Stylesheet _importedFrom;
    public Stylesheet _includedFrom;
    private Vector _includedStylesheets;
    private int _importPrecedence;
    private int _minimumDescendantPrecedence;
    private Map<String, Key> _keys;
    private SourceLoader _loader;
    private boolean _numberFormattingUsed;
    private boolean _simplified;
    private boolean _multiDocument;
    private boolean _callsNodeset;
    private boolean _hasIdCall;
    private boolean _templateInlining;
    private Output _lastOutputElement;
    private Properties _outputProperties;
    private int _outputMethod;
    public static final int UNKNOWN_OUTPUT = 0;
    public static final int XML_OUTPUT = 1;
    public static final int HTML_OUTPUT = 2;
    public static final int TEXT_OUTPUT = 3;
    
    public Stylesheet() {
        this._globals = new Vector();
        this._hasLocalParams = null;
        this._templates = new Vector();
        this._allValidTemplates = null;
        this._nextModeSerial = 1;
        this._modes = new HashMap();
        this._extensions = new HashMap();
        this._importedFrom = null;
        this._includedFrom = null;
        this._includedStylesheets = null;
        this._importPrecedence = 1;
        this._minimumDescendantPrecedence = -1;
        this._keys = new HashMap();
        this._loader = null;
        this._numberFormattingUsed = false;
        this._simplified = false;
        this._multiDocument = false;
        this._callsNodeset = false;
        this._hasIdCall = false;
        this._templateInlining = false;
        this._lastOutputElement = null;
        this._outputProperties = null;
        this._outputMethod = 0;
    }
    
    public int getOutputMethod() {
        return this._outputMethod;
    }
    
    private void checkOutputMethod() {
        if (this._lastOutputElement != null) {
            String method = this._lastOutputElement.getOutputMethod();
            if (method != null) {
                if (method.equals("xml")) {
                    this._outputMethod = 1;
                }
                else if (method.equals("html")) {
                    this._outputMethod = 2;
                }
                else if (method.equals("text")) {
                    this._outputMethod = 3;
                }
            }
        }
    }
    
    public boolean getTemplateInlining() {
        return this._templateInlining;
    }
    
    public void setTemplateInlining(boolean flag) {
        this._templateInlining = flag;
    }
    
    public boolean isSimplified() {
        return this._simplified;
    }
    
    public void setSimplified() {
        this._simplified = true;
    }
    
    public void setHasIdCall(boolean flag) {
        this._hasIdCall = flag;
    }
    
    public void setOutputProperty(String key, String value) {
        if (this._outputProperties == null) {
            this._outputProperties = new Properties();
        }
        this._outputProperties.setProperty(key, value);
    }
    
    public void setOutputProperties(Properties props) {
        this._outputProperties = props;
    }
    
    public Properties getOutputProperties() {
        return this._outputProperties;
    }
    
    public Output getLastOutputElement() {
        return this._lastOutputElement;
    }
    
    public void setMultiDocument(boolean flag) {
        this._multiDocument = flag;
    }
    
    public boolean isMultiDocument() {
        return this._multiDocument;
    }
    
    public void setCallsNodeset(boolean flag) {
        if (flag) {
            this.setMultiDocument(flag);
        }
        this._callsNodeset = flag;
    }
    
    public boolean callsNodeset() {
        return this._callsNodeset;
    }
    
    public void numberFormattingUsed() {
        this._numberFormattingUsed = true;
        Stylesheet parent = this.getParentStylesheet();
        if (null != parent) {
            parent.numberFormattingUsed();
        }
    }
    
    public void setImportPrecedence(int precedence) {
        this._importPrecedence = precedence;
        Iterator<SyntaxTreeNode> elements = this.elements();
        while (elements.hasNext()) {
            SyntaxTreeNode child = (SyntaxTreeNode)elements.next();
            if ((child instanceof Include)) {
                Stylesheet included = ((Include)child).getIncludedStylesheet();
                if (included != null && included._includedFrom == this) {
                    included.setImportPrecedence(precedence);
                }
                else {
                    continue;
                }
            }
            else {
                continue;
            }
        }
        if (this._importedFrom != null) {
            if (this._importedFrom.getImportPrecedence() < precedence) {
                Parser parser = this.getParser();
                int nextPrecedence = parser.getNextImportPrecedence();
                this._importedFrom.setImportPrecedence(nextPrecedence);
            }
        }
        else if (this._includedFrom != null && this._includedFrom.getImportPrecedence() != precedence) {
            this._includedFrom.setImportPrecedence(precedence);
        }
    }
    
    public int getImportPrecedence() {
        return this._importPrecedence;
    }
    
    public int getMinimumDescendantPrecedence() {
        if (this._minimumDescendantPrecedence == -1) {
            int min = this.getImportPrecedence();
            for (int inclImpCount = (this._includedStylesheets != null) ? this._includedStylesheets.size() : 0, i = 0; i < inclImpCount; ++i) {
                int prec = ((Stylesheet)this._includedStylesheets.elementAt(i)).getMinimumDescendantPrecedence();
                if (prec < min) {
                    min = prec;
                }
            }
            this._minimumDescendantPrecedence = min;
        }
        return this._minimumDescendantPrecedence;
    }
    
    public boolean checkForLoop(String systemId) {
        return (this._systemId != null && this._systemId.equals(systemId)) || (this._parentStylesheet != null && this._parentStylesheet.checkForLoop(systemId));
    }
    
    public void setParser(Parser parser) {
        super.setParser(parser);
        this._name = this.makeStylesheetName("__stylesheet_");
    }
    
    public void setParentStylesheet(Stylesheet parent) {
        this._parentStylesheet = parent;
    }
    
    public Stylesheet getParentStylesheet() {
        return this._parentStylesheet;
    }
    
    public void setImportingStylesheet(Stylesheet parent) {
        (this._importedFrom = parent).addIncludedStylesheet(this);
    }
    
    public void setIncludingStylesheet(Stylesheet parent) {
        (this._includedFrom = parent).addIncludedStylesheet(this);
    }
    
    public void addIncludedStylesheet(Stylesheet child) {
        if (this._includedStylesheets == null) {
            this._includedStylesheets = new Vector();
        }
        this._includedStylesheets.addElement(child);
    }
    
    public void setSystemId(String systemId) {
        if (systemId != null) {
            this._systemId = SystemIDResolver.getAbsoluteURI(systemId);
        }
    }
    
    public String getSystemId() {
        return this._systemId;
    }
    
    public void setSourceLoader(SourceLoader loader) {
        this._loader = loader;
    }
    
    public SourceLoader getSourceLoader() {
        return this._loader;
    }
    
    private QName makeStylesheetName(String prefix) {
        return this.getParser().getQName(prefix + this.getXSLTC().nextStylesheetSerial());
    }
    
    public boolean hasGlobals() {
        return this._globals.size() > 0;
    }
    
    public boolean hasLocalParams() {
        if (this._hasLocalParams == null) {
            Vector<Object> templates = this.getAllValidTemplates();
            int n = templates.size();
            int i = 0;
            while (i < n) {
                Template template = (Template)templates.elementAt(i);
                if (template.hasParams()) {
                    this._hasLocalParams = Boolean.TRUE;
                    return true;
                }
                ++i;
            }
            this._hasLocalParams = Boolean.FALSE;
            return false;
        }
        return (boolean)this._hasLocalParams;
    }
    
    @Override
    protected void addPrefixMapping(String prefix, String uri) {
        if (prefix.equals("") && uri.equals("http://www.w3.org/1999/xhtml")) {
            return;
        }
        super.addPrefixMapping(prefix, uri);
        return;
    }
    
    private void extensionURI(String prefixes, SymbolTable stable) {
        if (prefixes != null) {
            StringTokenizer tokens = new StringTokenizer(prefixes);
            while (tokens.hasMoreTokens()) {
                String prefix = tokens.nextToken();
                String uri = this.lookupNamespace(prefix);
                if (uri != null) {
                    this._extensions.put(uri, prefix);
                }
                else {
                    continue;
                }
            }
        }
    }
    
    public boolean isExtension(String uri) {
        return this._extensions.get(uri) != null;
    }
    
    public void declareExtensionPrefixes(Parser parser) {
        SymbolTable stable = parser.getSymbolTable();
        String extensionPrefixes = this.getAttribute("extension-element-prefixes");
        this.extensionURI(extensionPrefixes, stable);
    }
    
    @Override
    public void parseContents(Parser parser) {
        SymbolTable stable = parser.getSymbolTable();
        this.addPrefixMapping("xml", "http://www.w3.org/XML/1998/namespace");
        Stylesheet sheet = stable.addStylesheet(this._name, this);
        if (sheet != null) {
            ErrorMsg err = new ErrorMsg("MULTIPLE_STYLESHEET_ERR", this);
            parser.reportError(3, err);
        }
        if (this._simplified) {
            stable.excludeURI("http://www.w3.org/1999/XSL/Transform");
            Template template = new Template();
            template.parseSimplified(this, parser);
        }
        else {
            this.parseOwnChildren(parser);
        }
    }
    
    public final void parseOwnChildren(Parser parser) {
        SymbolTable stable = parser.getSymbolTable();
        String excludePrefixes = this.getAttribute("exclude-result-prefixes");
        String extensionPrefixes = this.getAttribute("extension-element-prefixes");
        stable.pushExcludedNamespacesContext();
        stable.excludeURI("http://www.w3.org/1999/XSL/Transform");
        stable.excludeNamespaces(excludePrefixes);
        stable.excludeNamespaces(extensionPrefixes);
        List<SyntaxTreeNode> contents = this.getContents();
        int count = contents.size();
        for (int i = 0; i < count; ++i) {
            SyntaxTreeNode child = (SyntaxTreeNode)contents.get(i);
            if (child instanceof VariableBase || child instanceof NamespaceAlias) {
                parser.getSymbolTable().setCurrentNode(child);
                child.parseContents(parser);
            }
        }
        for (int i = 0; i < count; ++i) {
            SyntaxTreeNode child = (SyntaxTreeNode)contents.get(i);
            if (!(child instanceof VariableBase) && !(child instanceof NamespaceAlias)) {
                parser.getSymbolTable().setCurrentNode(child);
                child.parseContents(parser);
            }
            if (!this._templateInlining && child instanceof Template) {
                Template template = (Template)child;
                String name = "template$dot$" + template.getPosition();
                template.setName(parser.getQName(name));
            }
        }
        stable.popExcludedNamespacesContext();
    }
    
    public void processModes() {
        if (this._defaultMode == null) {
            this._defaultMode = new Mode(null, this, "");
        }
        this._defaultMode.processPatterns(this._keys);
        for (Mode mode : this._modes.values()) {
            mode.processPatterns(this._keys);
        }
    }
    
    private void compileModes(ClassGenerator classGen) {
        this._defaultMode.compileApplyTemplates(classGen);
        for (Mode mode : this._modes.values()) {
            mode.compileApplyTemplates(classGen);
        }
    }
    
    public Mode getMode(QName modeName) {
        if (modeName == null) {
            if (this._defaultMode == null) {
                this._defaultMode = new Mode(null, this, "");
            }
            return this._defaultMode;
        }
        else {
            Mode mode = (Mode)this._modes.get(modeName.getStringRep());
            if (mode == null) {
                String suffix = Integer.toString(this._nextModeSerial++);
                this._modes.put(modeName.getStringRep(), mode = new Mode(modeName, this, suffix));
                return mode;
            }
            return mode;
        }
    }
    
    @Override
    public Type typeCheck(SymbolTable stable) throws TypeCheckError {
        for (int count = this._globals.size(), i = 0; i < count; ++i) {
            VariableBase var = (VariableBase)this._globals.elementAt(i);
            var.typeCheck(stable);
        }
        return this.typeCheckContents(stable);
    }
    
    @Override
    public void translate(ClassGenerator classGen, MethodGenerator methodGen) {
        this.translate();
    }
    
    private void addDOMField(ClassGenerator classGen) {
        FieldGen fgen = new FieldGen(1, Util.getJCRefType("Lcom/sun/org/apache/xalan/internal/xsltc/DOM;"), "_dom", classGen.getConstantPool());
        classGen.addField(fgen.getField());
    }
    
    private void addStaticField(ClassGenerator classGen, String type, String name) {
        FieldGen fgen = new FieldGen(12, Util.getJCRefType(type), name, classGen.getConstantPool());
        classGen.addField(fgen.getField());
    }
    
    public void translate() {
        this._className = this.getXSLTC().getClassName();
        ClassGenerator classGen = new ClassGenerator(this._className, "com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "", 33, null, this);
        this.addDOMField(classGen);
        this.compileTransform(classGen);
        Iterator<SyntaxTreeNode> elements = this.elements();
        while (elements.hasNext()) {
            SyntaxTreeNode element = (SyntaxTreeNode)elements.next();
            if ((element instanceof Template)) {
                Template template = (Template)element;
                this.getMode(template.getModeName()).addTemplate(template);
            }
            else if ((element instanceof AttributeSet)) {
                ((AttributeSet)element).translate(classGen, null);
            }
            else if ((element instanceof Output)) {
                Output output = (Output)element;
                if (output.enabled()) {
                    this._lastOutputElement = output;
                }
                else {
                    continue;
                }
            }
            else {
                continue;
            }
        }
        this.checkOutputMethod();
        this.processModes();
        this.compileModes(classGen);
        this.compileStaticInitializer(classGen);
        this.compileConstructor(classGen, this._lastOutputElement);
        if (!this.getParser().errorsFound()) {
            this.getXSLTC().dumpClass(classGen.getJavaClass());
        }
    }
    
    private void compileStaticInitializer(ClassGenerator classGen) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        InstructionList il = new InstructionList();
        MethodGenerator staticConst = new MethodGenerator(9, com.sun.org.apache.bcel.internal.generic.Type.VOID, null, null, "<clinit>", this._className, il, cpg);
        this.addStaticField(classGen, "[Ljava/lang/String;", "_sNamesArray");
        this.addStaticField(classGen, "[Ljava/lang/String;", "_sUrisArray");
        this.addStaticField(classGen, "[I", "_sTypesArray");
        this.addStaticField(classGen, "[Ljava/lang/String;", "_sNamespaceArray");
        for (int charDataFieldCount = this.getXSLTC().getCharacterDataCount(), i = 0; i < charDataFieldCount; ++i) {
            this.addStaticField(classGen, "[C", "_scharData" + i);
        }
        Vector<Object> namesIndex = this.getXSLTC().getNamesIndex();
        int size = namesIndex.size();
        String[] namesArray = new String[size];
        String[] urisArray = new String[size];
        int[] typesArray = new int[size];
        for (int j = 0; j < size; ++j) {
            String encodedName = (String)namesIndex.elementAt(j);
            int index;
            if ((index = encodedName.lastIndexOf(58)) > -1) {
                urisArray[j] = encodedName.substring(0, index);
            }
            ++index;
            if (encodedName.charAt(index) == '@') {
                typesArray[j] = 2;
                ++index;
            }
            else if (encodedName.charAt(index) == '?') {
                typesArray[j] = 13;
                ++index;
            }
            else {
                typesArray[j] = 1;
            }
            if (index == 0) {
                namesArray[j] = encodedName;
            }
            else {
                namesArray[j] = encodedName.substring(index);
            }
        }
        staticConst.markChunkStart();
        il.append(new PUSH(cpg, size));
        il.append(new ANEWARRAY(cpg.addClass("java.lang.String")));
        int namesArrayRef = cpg.addFieldref(this._className, "_sNamesArray", "[Ljava/lang/String;");
        il.append(new PUTSTATIC(namesArrayRef));
        staticConst.markChunkEnd();
        for (int k = 0; k < size; ++k) {
            String name = namesArray[k];
            staticConst.markChunkStart();
            il.append(new GETSTATIC(namesArrayRef));
            il.append(new PUSH(cpg, k));
            il.append(new PUSH(cpg, name));
            il.append(Stylesheet.AASTORE);
            staticConst.markChunkEnd();
        }
        staticConst.markChunkStart();
        il.append(new PUSH(cpg, size));
        il.append(new ANEWARRAY(cpg.addClass("java.lang.String")));
        int urisArrayRef = cpg.addFieldref(this._className, "_sUrisArray", "[Ljava/lang/String;");
        il.append(new PUTSTATIC(urisArrayRef));
        staticConst.markChunkEnd();
        for (int l = 0; l < size; ++l) {
            String uri = urisArray[l];
            staticConst.markChunkStart();
            il.append(new GETSTATIC(urisArrayRef));
            il.append(new PUSH(cpg, l));
            il.append(new PUSH(cpg, uri));
            il.append(Stylesheet.AASTORE);
            staticConst.markChunkEnd();
        }
        staticConst.markChunkStart();
        il.append(new PUSH(cpg, size));
        il.append(new NEWARRAY(BasicType.INT));
        int typesArrayRef = cpg.addFieldref(this._className, "_sTypesArray", "[I");
        il.append(new PUTSTATIC(typesArrayRef));
        staticConst.markChunkEnd();
        for (int m = 0; m < size; ++m) {
            int nodeType = typesArray[m];
            staticConst.markChunkStart();
            il.append(new GETSTATIC(typesArrayRef));
            il.append(new PUSH(cpg, m));
            il.append(new PUSH(cpg, nodeType));
            il.append(Stylesheet.IASTORE);
        }
        Vector<Object> namespaces = this.getXSLTC().getNamespaceIndex();
        staticConst.markChunkStart();
        il.append(new PUSH(cpg, namespaces.size()));
        il.append(new ANEWARRAY(cpg.addClass("java.lang.String")));
        int namespaceArrayRef = cpg.addFieldref(this._className, "_sNamespaceArray", "[Ljava/lang/String;");
        il.append(new PUTSTATIC(namespaceArrayRef));
        staticConst.markChunkEnd();
        for (int i2 = 0; i2 < namespaces.size(); ++i2) {
            String ns = (String)namespaces.elementAt(i2);
            staticConst.markChunkStart();
            il.append(new GETSTATIC(namespaceArrayRef));
            il.append(new PUSH(cpg, i2));
            il.append(new PUSH(cpg, ns));
            il.append(Stylesheet.AASTORE);
            staticConst.markChunkEnd();
        }
        int charDataCount = this.getXSLTC().getCharacterDataCount();
        int toCharArray = cpg.addMethodref("java.lang.String", "toCharArray", "()[C");
        for (int i3 = 0; i3 < charDataCount; ++i3) {
            staticConst.markChunkStart();
            il.append(new PUSH(cpg, this.getXSLTC().getCharacterData(i3)));
            il.append(new INVOKEVIRTUAL(toCharArray));
            il.append(new PUTSTATIC(cpg.addFieldref(this._className, "_scharData" + i3, "[C")));
            staticConst.markChunkEnd();
        }
        il.append(Stylesheet.RETURN);
        classGen.addMethod(staticConst);
    }
    
    private void compileConstructor(ClassGenerator classGen, Output output) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        InstructionList il = new InstructionList();
        MethodGenerator constructor = new MethodGenerator(1, com.sun.org.apache.bcel.internal.generic.Type.VOID, null, null, "<init>", this._className, il, cpg);
        il.append(classGen.loadTranslet());
        il.append(new INVOKESPECIAL(cpg.addMethodref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "<init>", "()V")));
        constructor.markChunkStart();
        il.append(classGen.loadTranslet());
        il.append(new GETSTATIC(cpg.addFieldref(this._className, "_sNamesArray", "[Ljava/lang/String;")));
        il.append(new PUTFIELD(cpg.addFieldref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "namesArray", "[Ljava/lang/String;")));
        il.append(classGen.loadTranslet());
        il.append(new GETSTATIC(cpg.addFieldref(this._className, "_sUrisArray", "[Ljava/lang/String;")));
        il.append(new PUTFIELD(cpg.addFieldref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "urisArray", "[Ljava/lang/String;")));
        constructor.markChunkEnd();
        constructor.markChunkStart();
        il.append(classGen.loadTranslet());
        il.append(new GETSTATIC(cpg.addFieldref(this._className, "_sTypesArray", "[I")));
        il.append(new PUTFIELD(cpg.addFieldref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "typesArray", "[I")));
        constructor.markChunkEnd();
        constructor.markChunkStart();
        il.append(classGen.loadTranslet());
        il.append(new GETSTATIC(cpg.addFieldref(this._className, "_sNamespaceArray", "[Ljava/lang/String;")));
        il.append(new PUTFIELD(cpg.addFieldref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "namespaceArray", "[Ljava/lang/String;")));
        constructor.markChunkEnd();
        constructor.markChunkStart();
        il.append(classGen.loadTranslet());
        il.append(new PUSH(cpg, 101));
        il.append(new PUTFIELD(cpg.addFieldref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "transletVersion", "I")));
        constructor.markChunkEnd();
        if (this._hasIdCall) {
            constructor.markChunkStart();
            il.append(classGen.loadTranslet());
            il.append(new PUSH(cpg, Boolean.TRUE));
            il.append(new PUTFIELD(cpg.addFieldref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "_hasIdCall", "Z")));
            constructor.markChunkEnd();
        }
        if (output != null) {
            constructor.markChunkStart();
            output.translate(classGen, constructor);
            constructor.markChunkEnd();
        }
        if (this._numberFormattingUsed) {
            constructor.markChunkStart();
            DecimalFormatting.translateDefaultDFS(classGen, constructor);
            constructor.markChunkEnd();
        }
        il.append(Stylesheet.RETURN);
        classGen.addMethod(constructor);
    }
    
    private String compileTopLevel(ClassGenerator classGen) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        com.sun.org.apache.bcel.internal.generic.Type[] argTypes = { Util.getJCRefType("Lcom/sun/org/apache/xalan/internal/xsltc/DOM;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;") };
        String[] argNames = { "document", "iterator", "handler" };
        InstructionList il = new InstructionList();
        MethodGenerator toplevel = new MethodGenerator(1, com.sun.org.apache.bcel.internal.generic.Type.VOID, argTypes, argNames, "topLevel", this._className, il, classGen.getConstantPool());
        toplevel.addException("com.sun.org.apache.xalan.internal.xsltc.TransletException");
        LocalVariableGen current = toplevel.addLocalVariable("current", com.sun.org.apache.bcel.internal.generic.Type.INT, null, null);
        int setFilter = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "setFilter", "(Lcom/sun/org/apache/xalan/internal/xsltc/StripFilter;)V");
        int gitr = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "getIterator", "()Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;");
        il.append(toplevel.loadDOM());
        il.append(new INVOKEINTERFACE(gitr, 1));
        il.append(toplevel.nextNode());
        current.setStart(il.append(new ISTORE(current.getIndex())));
        Vector varDepElements = new Vector(this._globals);
        Iterator<SyntaxTreeNode> elements = this.elements();
        while (elements.hasNext()) {
            SyntaxTreeNode element = (SyntaxTreeNode)elements.next();
            if ((element instanceof Key)) {
                varDepElements.add(element);
            }
            else {
                continue;
            }
        }
        varDepElements = this.resolveDependencies(varDepElements);
        for (int count = varDepElements.size(), i = 0; i < count; ++i) {
            TopLevelElement tle = (TopLevelElement)varDepElements.elementAt(i);
            tle.translate(classGen, toplevel);
            if ((tle instanceof Key)) {
                Key key = (Key)tle;
                this._keys.put(key.getName(), key);
            }
        }
        Vector<Object> whitespaceRules = new Vector();
        elements = this.elements();
        while (elements.hasNext()) {
            SyntaxTreeNode element2 = (SyntaxTreeNode)elements.next();
            if ((element2 instanceof DecimalFormatting)) {
                ((DecimalFormatting)element2).translate(classGen, toplevel);
            }
            else if ((element2 instanceof Whitespace)) {
                whitespaceRules.addAll(((Whitespace)element2).getRules());
            }
            else {
                continue;
            }
        }
        if (whitespaceRules.size() > 0) {
            Whitespace.translateRules(whitespaceRules, classGen);
        }
        if (classGen.containsMethod("stripSpace", "(Lcom/sun/org/apache/xalan/internal/xsltc/DOM;II)Z") != null) {
            il.append(toplevel.loadDOM());
            il.append(classGen.loadTranslet());
            il.append(new INVOKEINTERFACE(setFilter, 2));
        }
        il.append(Stylesheet.RETURN);
        classGen.addMethod(toplevel);
        return "(Lcom/sun/org/apache/xalan/internal/xsltc/DOM;Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;)V";
    }
    
    private Vector resolveDependencies(Vector input) {
        Vector<TopLevelElement> result = new Vector();
        while (input.size() > 0) {
            boolean changed = false;
            int i = 0;
            while (i < input.size()) {
                TopLevelElement vde = (TopLevelElement)input.elementAt(i);
                Vector<Object> dep = vde.getDependencies();
                if (dep == null || result.containsAll(dep)) {
                    result.addElement(vde);
                    input.remove(i);
                    changed = true;
                }
                else {
                    ++i;
                }
            }
            if (!changed) {
                ErrorMsg err = new ErrorMsg("CIRCULAR_VARIABLE_ERR", input.toString(), this);
                this.getParser().reportError(3, err);
                return result;
            }
            continue;
        }
        return result;
    }
    
    private String compileBuildKeys(ClassGenerator classGen) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        com.sun.org.apache.bcel.internal.generic.Type[] argTypes = { Util.getJCRefType("Lcom/sun/org/apache/xalan/internal/xsltc/DOM;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;"), com.sun.org.apache.bcel.internal.generic.Type.INT };
        String[] argNames = { "document", "iterator", "handler", "current" };
        InstructionList il = new InstructionList();
        MethodGenerator buildKeys = new MethodGenerator(1, com.sun.org.apache.bcel.internal.generic.Type.VOID, argTypes, argNames, "buildKeys", this._className, il, classGen.getConstantPool());
        buildKeys.addException("com.sun.org.apache.xalan.internal.xsltc.TransletException");
        Iterator<SyntaxTreeNode> elements = this.elements();
        while (elements.hasNext()) {
            SyntaxTreeNode element = (SyntaxTreeNode)elements.next();
            if ((element instanceof Key)) {
                Key key = (Key)element;
                key.translate(classGen, buildKeys);
                this._keys.put(key.getName(), key);
            }
            else {
                continue;
            }
        }
        il.append(Stylesheet.RETURN);
        buildKeys.stripAttributes(true);
        buildKeys.setMaxLocals();
        buildKeys.setMaxStack();
        buildKeys.removeNOPs();
        classGen.addMethod(buildKeys.getMethod());
        return "(Lcom/sun/org/apache/xalan/internal/xsltc/DOM;Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;I)V";
    }
    
    private void compileTransform(ClassGenerator classGen) {
        ConstantPoolGen cpg = classGen.getConstantPool();
        com.sun.org.apache.bcel.internal.generic.Type[] argTypes = { Util.getJCRefType("Lcom/sun/org/apache/xalan/internal/xsltc/DOM;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;"), Util.getJCRefType("Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;") };
        String[] argNames = { "document", "iterator", "handler" };
        InstructionList il = new InstructionList();
        MethodGenerator transf = new MethodGenerator(1, com.sun.org.apache.bcel.internal.generic.Type.VOID, argTypes, argNames, "transform", this._className, il, classGen.getConstantPool());
        transf.addException("com.sun.org.apache.xalan.internal.xsltc.TransletException");
        int check = cpg.addMethodref("com.sun.org.apache.xalan.internal.xsltc.runtime.BasisLibrary", "resetPrefixIndex", "()V");
        il.append(new INVOKESTATIC(check));
        LocalVariableGen current = transf.addLocalVariable("current", com.sun.org.apache.bcel.internal.generic.Type.INT, null, null);
        String applyTemplatesSig = classGen.getApplyTemplatesSig();
        int applyTemplates = cpg.addMethodref(this.getClassName(), "applyTemplates", applyTemplatesSig);
        int domField = cpg.addFieldref(this.getClassName(), "_dom", "Lcom/sun/org/apache/xalan/internal/xsltc/DOM;");
        il.append(classGen.loadTranslet());
        if (this.isMultiDocument()) {
            il.append(new NEW(cpg.addClass("com.sun.org.apache.xalan.internal.xsltc.dom.MultiDOM")));
            il.append(Stylesheet.DUP);
        }
        il.append(classGen.loadTranslet());
        il.append(transf.loadDOM());
        il.append(new INVOKEVIRTUAL(cpg.addMethodref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "makeDOMAdapter", "(Lcom/sun/org/apache/xalan/internal/xsltc/DOM;)Lcom/sun/org/apache/xalan/internal/xsltc/dom/DOMAdapter;")));
        if (this.isMultiDocument()) {
            int init = cpg.addMethodref("com.sun.org.apache.xalan.internal.xsltc.dom.MultiDOM", "<init>", "(Lcom/sun/org/apache/xalan/internal/xsltc/DOM;)V");
            il.append(new INVOKESPECIAL(init));
        }
        il.append(new PUTFIELD(domField));
        int gitr = cpg.addInterfaceMethodref("com.sun.org.apache.xalan.internal.xsltc.DOM", "getIterator", "()Lcom/sun/org/apache/xml/internal/dtm/DTMAxisIterator;");
        il.append(transf.loadDOM());
        il.append(new INVOKEINTERFACE(gitr, 1));
        il.append(transf.nextNode());
        current.setStart(il.append(new ISTORE(current.getIndex())));
        il.append(classGen.loadTranslet());
        il.append(transf.loadHandler());
        int index = cpg.addMethodref("com.sun.org.apache.xalan.internal.xsltc.runtime.AbstractTranslet", "transferOutputSettings", "(Lcom/sun/org/apache/xml/internal/serializer/SerializationHandler;)V");
        il.append(new INVOKEVIRTUAL(index));
        String keySig = this.compileBuildKeys(classGen);
        int keyIdx = cpg.addMethodref(this.getClassName(), "buildKeys", keySig);
        Iterator<SyntaxTreeNode> toplevel = this.elements();
        if (this._globals.size() > 0 || toplevel.hasNext()) {
            String topLevelSig = this.compileTopLevel(classGen);
            int topLevelIdx = cpg.addMethodref(this.getClassName(), "topLevel", topLevelSig);
            il.append(classGen.loadTranslet());
            il.append(classGen.loadTranslet());
            il.append(new GETFIELD(domField));
            il.append(transf.loadIterator());
            il.append(transf.loadHandler());
            il.append(new INVOKEVIRTUAL(topLevelIdx));
        }
        il.append(transf.loadHandler());
        il.append(transf.startDocument());
        il.append(classGen.loadTranslet());
        il.append(classGen.loadTranslet());
        il.append(new GETFIELD(domField));
        il.append(transf.loadIterator());
        il.append(transf.loadHandler());
        il.append(new INVOKEVIRTUAL(applyTemplates));
        il.append(transf.loadHandler());
        il.append(transf.endDocument());
        il.append(Stylesheet.RETURN);
        classGen.addMethod(transf);
    }
    
    private void peepHoleOptimization(MethodGenerator methodGen) {
        String pattern = "`aload'`pop'`instruction'";
        InstructionList il = methodGen.getInstructionList();
        InstructionFinder find = new InstructionFinder(il);
        Iterator<InstructionHandle> iter = find.search("`aload'`pop'`instruction'");
        while (iter.hasNext()) {
            InstructionHandle[] match = (InstructionHandle[])(InstructionHandle[])iter.next();
            try {
                il.delete(match[0], match[1]);
            }
            catch (TargetLostException ex) {}
        }
    }
    
    public int addParam(Param param) {
        this._globals.addElement(param);
        return this._globals.size() - 1;
    }
    
    public int addVariable(Variable global) {
        this._globals.addElement(global);
        return this._globals.size() - 1;
    }
    
    @Override
    public void display(int indent) {
        this.indent(indent);
        Util.println("Stylesheet");
        this.displayContents(indent + 4);
    }
    
    public String getNamespace(String prefix) {
        return this.lookupNamespace(prefix);
    }
    
    public String getClassName() {
        return this._className;
    }
    
    public Vector getTemplates() {
        return this._templates;
    }
    
    public Vector getAllValidTemplates() {
        if (this._includedStylesheets == null) {
            return this._templates;
        }
        if (this._allValidTemplates == null) {
            Vector<Object> templates = new Vector();
            templates.addAll(this._templates);
            for (int size = this._includedStylesheets.size(), i = 0; i < size; ++i) {
                Stylesheet included = (Stylesheet)this._includedStylesheets.elementAt(i);
                templates.addAll(included.getAllValidTemplates());
            }
            if (this._parentStylesheet != null) {
                return templates;
            }
            this._allValidTemplates = templates;
        }
        return this._allValidTemplates;
    }
    
    protected void addTemplate(Template template) {
        this._templates.addElement(template);
    }
}
